<?php

namespace FiloBlu\ExtInventory\Model\ShippingAlgorithms\SourceSlots;

use Magento\InventorySourceSelectionApi\Api\Data\InventoryRequestInterface;
use Magento\InventorySourceSelectionApi\Api\Data\SourceSelectionResultInterface;
use Magento\InventorySourceSelectionApi\Api\Data\SourceSelectionResultInterfaceFactory;
use Magento\InventorySourceSelectionApi\Model\GetInStockSourceItemsBySkusAndSortedSource;
use Magento\InventorySourceSelectionApi\Model\GetSourceItemQtyAvailableInterface;
use Magento\InventorySourceSelectionApi\Model\SourceSelectionInterface;
use Magento\InventorySourceSelectionApi\Model\Algorithms\Result\GetDefaultSortedSourcesResult;
use Magento\Framework\App\ObjectManager;
use Magento\InventoryApi\Api\GetSourceItemsBySkuInterface;
use FiloBlu\ExtInventory\Model\ShippingAlgorithms\SourceSlots\CalculationsCore;

use Magento\InventorySourceSelectionApi\Api\Data\SourceSelectionItemInterfaceFactory;

use FiloBlu\ExtInventory\Model\ShippingAlgorithms\SourceSlots\SourcesData;

class Algorithm implements SourceSelectionInterface
{

    //namespace Magento\InventorySourceSelectionApi\Model\Algorithms\Result;
    //class GetDefaultSortedSourcesResult

    private $getDefaultSortedSourcesResult;
    /**
     * @var \FiloBlu\ExtInventory\Model\ShippingAlgorithms\SourceSlots\SourcesData
     */
    private $sourcesData;
    private $getSourceItemsBySkuInterface;
    protected $calculationsCore;

    protected $slots;

    private $getInStockSourceItemsBySkusAndSortedSource;
    private $sourceSelectionItemFactory;
    private $getSourceItemQtyAvailable;
    private $sourceSelectionResultFactory;

    public function __construct(
        SourcesData $sourcesData,
        GetSourceItemsBySkuInterface $getSourceItemsBySkuInterface,
        CalculationsCore $calculationsCore,
        SourceSelectionResultInterfaceFactory $sourceSelectionResultFactory,
        SourceSelectionItemInterfaceFactory $sourceSelectionItemFactory,
        GetDefaultSortedSourcesResult $getDefaultSortedSourcesResult = null,
        GetInStockSourceItemsBySkusAndSortedSource $getInStockSourceItemsBySkusAndSortedSource = null,
        GetSourceItemQtyAvailableInterface $getSourceItemQtyAvailable = null
    ) {
        $this->sourcesData = $sourcesData;
        $this->getSourceItemsBySkuInterface = $getSourceItemsBySkuInterface;
        $this->calculationsCore = $calculationsCore;
        $this->sourceSelectionResultFactory = $sourceSelectionResultFactory;
        $this->sourceSelectionItemFactory = $sourceSelectionItemFactory;
        $this->getDefaultSortedSourcesResult = $getDefaultSortedSourcesResult ?: ObjectManager::getInstance()->get(GetDefaultSortedSourcesResult::class);
        $this->getInStockSourceItemsBySkusAndSortedSource = $getInStockSourceItemsBySkusAndSortedSource ?: ObjectManager::getInstance()->get(GetInStockSourceItemsBySkusAndSortedSource::class);
        $this->getSourceItemQtyAvailable = $getSourceItemQtyAvailable ?? ObjectManager::getInstance()->get(GetSourceItemQtyAvailableInterface::class);
    }

    public function execute(InventoryRequestInterface $inventoryRequest): SourceSelectionResultInterface
    {
        // Stock ID of table inventory_stock
        $stockId = $inventoryRequest->getStockId();

        $this->setSourcesSlots();

        $itemsToDeliver = [];
        foreach ($inventoryRequest->getItems() as $item) {
            $itemsToDeliver[$item->getSku()] = $item->getQty();
        }

        $combinationsCalculated = $this->calculationsCore->calculate($itemsToDeliver,$stockId);

        $result = $this->buildResult($combinationsCalculated,$itemsToDeliver,$stockId);

        return $result;
    }

    /*
     * TODO: Dynamically set $stockId from $order
     */
    public function getResultPreviewFromOrder(\Magento\Sales\Api\Data\OrderInterface $order, $stockId = 2){
        $this->setSourcesSlots();

        $itemsToDeliver = [];
        foreach ($order->getAllVisibleItems() as $item) {
            $itemsToDeliver[$item->getSku()] = $item->getQtyOrdered();
        }

        $combinationsCalculated = $this->calculationsCore->calculate($itemsToDeliver,$stockId);

        return $combinationsCalculated;
    }

    private function setSourcesSlots(){
        // Reading the slots configuration of the sources
        $occupiedSlots = $this->sourcesData->getSourcesOccupiedSlotsNumber();
        $maxSlots = $this->sourcesData->getSourcesDailyMaxSlots();
        $this->slots = array_merge_recursive($occupiedSlots,$maxSlots);
    }

    // Result must be "SourceSelectionResultInterfaceFactory"
    // Example: Magento\InventorySourceSelectionApi\Model\Algorithms\Result\GetDefaultSortedSourcesResult
    public function buildResult($combinationsCalculated,$itemsToDeliver,$stockId){
        $sortedSourceCodes = $this->sourcesData->getEnabledSourcesCodesArrayOrderedByPriorityByStockId($stockId);

        $sourceItems =
            $this->getInStockSourceItemsBySkusAndSortedSource->execute(
                array_keys($itemsToDeliver),
                $sortedSourceCodes
            );

        $sourceItemSelections = [];

        foreach ($sourceItems as $sourceItem) {
            $sourceItemQtyAvailable = $this->getSourceItemQtyAvailable->execute($sourceItem);
            $quantityToDeliverCalculated = $combinationsCalculated[0][$sourceItem->getSourceCode()][$sourceItem->getSku()];
            $qtyToDeduct = min($sourceItemQtyAvailable, $quantityToDeliverCalculated ?? 0.0);

            $sourceItemSelections[] = $this->sourceSelectionItemFactory->create([
                'sourceCode' => $sourceItem->getSourceCode(),
                'sku' => $sourceItem->getSku(),
                'qtyToDeduct' => $qtyToDeduct,
                'qtyAvailable' => $sourceItemQtyAvailable
            ]);

            $itemsToDeliver[$sourceItem->getSku()] -= $qtyToDeduct;
        }

        $isShippable = false;
        foreach ($itemsToDeliver as $itemToDeliver) {
            if (!$this->isZero($itemToDeliver)) {
                $isShippable = false;
                break;
            }
        }

        return $this->sourceSelectionResultFactory->create(
            [
                'sourceItemSelections' => $sourceItemSelections,
                'isShippable' => $isShippable
            ]
        );
    }

    private function isZero(float $floatNumber): bool
    {
        return $floatNumber < 0.0000001;
    }


}